program DataSaltDim

  use sep
  use smooth_mod
  implicit none


  character(len=128)  :: outfile1 
  integer             :: nt, nx, slen
  real                :: dt, dx, ot, ox
  real                :: val, thres, ampmax, bound, gx
  integer             :: it, ix, imax, igx, ibound
  real, allocatable, dimension(:) :: trace, gxwgt

!! velocity values
  real, dimension(:,:), pointer :: wgtarray,  arrayin



!! Open I/O
  call sep_init()


!! Read file dimensions (velocity model with salt)
  call from_history("n1", nt);  call from_history("o1", ot);  call from_history("d1", dt)
  call from_history("n2", nx);  call from_history("o2", ox);  call from_history("d2", dx)

!! Read threshold value
  call from_param("thres", thres, 0.99);
  call from_param("slen", slen, 10);
  call from_param("xsrc", gx, 9999.) 
 
  !! Read in filename
  call from_param('outfile1',outfile1);  
  igx = (gx-ox)/dx+1.
 

  write(0,*) '######################################'
  write(0,*) 'Program: createveltig.x'
  write(0,*) 'nt, ot, dt',  nt, ot, dt
  write(0,*) 'nx, ox, dx',  nx, ox, dx
  write(0,*) 'thres, gx', thres, gx, igx
  write(0,*) '######################################'

    !! Output files
    call auxputch("n1",'i',nt,outfile1);        call auxputch("n2",'i',nx,outfile1);
    call auxputch("d1",'f',dt,outfile1);        call auxputch("d2",'f',dx,outfile1)
    call auxputch("o1",'f',ot,outfile1);        call auxputch("o2",'f',ox,outfile1)
    call auxputch("label1",'s','t(s)',outfile1);     call auxputch("label2",'s','x(m)',outfile1) 


  call sep_close()

  allocate(wgtarray(nt,nx),arrayin(nt,nx))
  allocate(trace(nt),gxwgt(nx))
 
  call sep_read(arrayin)

  do ix=1,nx
     do it=1,nt
        wgtarray(it,ix) = 1;
     end do
  end do
  
  ibound = 150

  do ix=1,nx
     if ( (abs(ix-igx) .lt. ibound) ) then
        gxwgt(ix) = min(real(abs(igx-ix))/real(ibound)+0.01,1.)	
     else 
        gxwgt(ix) = 1.
     end if
     if (mod(ix,2) == 1)  write(0,*) ix, gxwgt(ix)
  end do

   ampmax = maxval(arrayin)
  write(0,*) 'ampmax', ampmax
    

  ! Now all the salt in
  do ix=1,nx
     trace(1:nt) = arrayin(1:nt,ix)
     bound = thres*maxval(trace)*gxwgt(ix)
     do it=1,nt
        val = abs(arrayin(it,ix))
        if (val .gt. bound) then
           wgtarray(it,ix) = 0.0
        end if
     end do
  end do  

   
  !  smooth wgtarray
  call triangle2(slen,slen,nt,nx,wgtarray,wgtarray)
 
   ! Now all the salt in
  do ix=1,nx
     do it=1,nt
        val = abs(wgtarray(it,ix))
        if (val .lt. 1.) then
           wgtarray(it,ix) = 0.1
	else
	   wgtarray(it,ix) = 1.
        end if
    end do
  end do 

  do ix=1,nx
     trace(1:nt) = arrayin(1:nt,ix)
     imax = maxloc(trace,1)
     wgtarray(1:imax,ix) =0.
     write(0,*) 'imax',imax
  end do


  call triangle2(20,20,nt,nx,wgtarray,wgtarray)




!! Write new vel values to output
  call sep_write(wgtarray,outfile1);



  call exit(0)
   
end program 
