!This program creates synthetic data
!Initial version written by Noha in September 2012
!Modified by Mandy in November 2012


!Synopsis < in=velocity aux=source.H  > recdata.H
! dt_data=[real] nt_data=[int] : time-axis of data  
! xsrc=[real]  zsrc=[real] : Source position
! nrx=[int], orx=[real], drx=[real]  : Receiver positions


Program Synth_data

  use sep
  use prop_mod
  implicit none
  
  logical(4) :: success
  logical   :: fixspread, wfldmovie
  real :: ot_data, dt_data,oz,ox,dz,dx,xsrc,zsrc,ost,dst,drx,orx, vmax,tdelay
  integer :: npad,nt_data,nst,nnx, nnz, nx,nz,nrx,nop
  real, allocatable, dimension(:,:) :: vel,dat
  real, allocatable, dimension(:) :: wavelet
  character(len=128)              :: souWfile_, imgWfile_, recWfile_, outfile_
  
  call sep_init()
  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  !Program read-ins
  !velocity model is entered as standard input
  call from_history("n1", nz)
  call from_history("o1", oz)
  call from_history("d1", dz)
  call from_history("n2", nx)
  call from_history("o2", ox)
  call from_history("d2", dx)
  
  !the auxiliary file is the source wavelet
  call from_aux("aux", "n1", nst)
  call from_aux("aux", "o1", ost)
  call from_aux("aux", "d1", dst)

  ! read source location 
  call from_param("xsrc",xsrc,(ox+(nx-1)*dx)/2.) 
  call from_param("zsrc",zsrc,dz/2.) 

  ! read data parameter
  call from_param("dt_data",dt_data,0.004)
  call from_param("nt_data",nt_data,500)
  call from_param("ot_data",ot_data,0.)
  call from_param("fixspread",fixspread,.true.);
  call from_param("tdelay",tdelay)

  call from_param("npad",npad,30)
  nnz=nz
  nnx=nx
  nz = nnz+2*npad
  nx = nnx+2*npad


  ! rather or not to output source wavefield movie
  call from_param("wfldout",wfldmovie,.false.)
  call from_param("souwfld",souWfile_,"souwfld.H")
  call from_param("imgwfld",imgWfile_,"imgwfld.H")  
  call from_param("recwfld",recWfile_,"recwfld.H")
  call from_param("outfile",outfile_)


  call from_param("vmax",vmax)

  if (fixspread) then
     nop=4
     nrx= nx-2*max(npad,nop)
     drx=dx
     orx=ox
  else
     call from_param("nrx", nrx); 
     call from_param("drx", drx); 
     call from_param("orx",orx)
     orx = orx + xsrc
  end if

!  if (nt_data .ne. nst) call erexit('nst should equal to nt_data')
  if (dst .ne. dt_data) call erexit('dst should equal to dt_data')


  ! Memory allocation
  allocate (vel(nz,nx),wavelet(nt_data), dat(nt_data,nrx))

  ! Read velocity model
  call sep_read(vel(npad+1:nz-npad,npad+1:nx-npad));  call from_param("vmax",vmax);
  ! Read source wavelet
  wavelet=0.
  call sep_read(wavelet,"aux")
 
   write(0,*) 'tdelay', tdelay

  ! Initiate propagation
  call init_propagate(vel,oz,ox,dz,dx,dt_data,nt_data,wfldmovie,souWfile_,recWfile_,imgWfile_,vmax,npad) 

  ! Initiate source parameters
  call source_init(wavelet,xsrc,zsrc,tdelay)
  
  ! Initiate receiver parameters
  call receiver_init(.false.,zsrc,xsrc,orx,drx,dat)
  
  ! do the propagation
  call model_shot(dat) 
  
  ! Output data
    call auxputch("n1",'i',nt_data,outfile_)
    call auxputch("n2",'i',nrx,outfile_)
    call auxputch("d1",'f',dt_data,outfile_)
    call auxputch("d2",'f',drx,outfile_)
    call auxputch("o1",'f',ot_data,outfile_)
    call auxputch("o2",'f',orx,outfile_)
    call auxputch("label1",'s','t(s)',outfile_) 
    call auxputch("label2",'s','rx(m)',outfile_) 
  call sep_write(dat,outfile_)

  call sep_close()

end program
