!Born Modelling code
!Modified by Mandy in November 2012

Program bornmodel

  use sep
  use prop_mod
  implicit none

  logical(4) :: success
  logical                               :: fixspread, wfldmovie,VDer, DW
  real                                  :: dx, dz, ox, oz, ost, dst, orx, drx, ot_data, dt_data
  real                                  :: xsrc, zsrc, vmax,tdelay
  integer                               :: nnx, nnz, nx, nz, nst, nt_data, nrx, nop, npad
  integer                               :: nix, niz, iz, i, ix, it
  real                                  :: dix, oix, diz, oiz
  real, allocatable, dimension(:,:)   :: vel, image, mwarr, dat, DWarray
  real, allocatable, dimension(:)   ::wavelet, imgtaper
  character(len=128)              :: souWfile_, imgWfile_, recWfile_, outfile_

  call sep_init()

  ! read in velocity model
  call from_history("n1", nz)
  call from_history("o1", oz)
  call from_history("d1", dz)
  call from_history("n2", nx)
  call from_history("o2", ox)
  call from_history("d2", dx)

  ! read source wavelet information
  call from_aux("aux", "n1", nst)
  call from_aux("aux", "o1", ost)
  call from_aux("aux", "d1", dst)

  ! read rtm image file info
  call from_aux("imgfile", "n1", niz)
  call from_aux("imgfile", "o1", oiz)
  call from_aux("imgfile", "d1", diz)
  call from_aux("imgfile", "n2", nix)
  call from_aux("imgfile", "o2", oix)    
  call from_aux("imgfile", "d2", dix)

  write(0,*) 'nz, oz, dz', nz, oz, dz
  write(0,*) 'nx, ox, dx', nx, ox, dx
  write(0,*) 'niz, oiz, diz', niz, oiz, diz
  write(0,*) 'nix, oix, dix', nix, oix, dix
 
  ! read source location
  call from_param("zsrc",zsrc)
  call from_param("xsrc",xsrc)
  call from_param("npad",npad,30)
  


  ! read data parameter
  call from_param("dt_data",dt_data,0.004)
  call from_param("nt_data",nt_data,500)
  call from_param("ot_data",ot_data,0.)
  call from_param("fixspread",fixspread,.true.);
  call from_param("tdelay",tdelay);

 
  ! Read wavefield movie switch
  call from_param("wfldout",wfldmovie,.false.)
  call from_param("VDer",VDer,.false.)
  call from_param("DW",DW,.false.)
  call from_param("souwfld",souWfile_,"souwfld.H")
  call from_param("imgwfld",imgWfile_,"imgwfld.H")  
  call from_param("recwfld",recWfile_,"recwfld.H")
  call from_param("outfile",outfile_)



  ! Error check
  if ( (niz .ne. nz) .or. (nix .ne. nx) ) call erexit('Image and velocity files dimension mismatch')
  if ( (diz .ne. dz) .or. (dix .ne. dx) ) call erexit('Image and velocity files dimension mismatch')
  if ( (oiz .ne. oz) .or. (oix .ne. ox) ) call erexit('Image and velocity files dimension mismatch')
  if (dst .ne. dt_data) call erexit('dst should equal to dt_data')


  nnz=nz
  nnx=nx
  nz = nnz+2*npad
  nx = nnx+2*npad
  if (fixspread) then
     nop=4
     nrx= nx-2*max(npad,nop)
     drx=dx
     orx=ox
  else
     call from_param("nrx", nrx); 
     call from_param("drx", drx); 
     call from_param("orx",orx)
     orx = orx + xsrc
  end if



  ! Allocate memory
  allocate(vel(nz,nx),dat(nt_data,nrx),image(nz,nx),wavelet(nt_data))

  ! Read velocity model
  call sep_read(vel(npad+1:nz-npad,npad+1:nx-npad));  call from_param("vmax",vmax);

  ! Read source wavelet
  call sep_read(wavelet,"aux")
  
  ! Read data
  call sep_read(image(npad+1:nz-npad,npad+1:nx-npad),"imgfile")

  ! Apply model weighting
  allocate(mwarr(nnz,nnx))
  mwarr=1.
  call sep_read(mwarr(1:nnz,1:nnx),"mwfile")
  do ix=npad+1,nx-npad
     do iz=npad+1,nz-npad
        image(iz,ix) = image(iz,ix)*mwarr(iz-npad,ix-npad)
     end do
  end do

   ! Apply filter to model
    if (VDer) then
       call Ver2ndDer(.false.,image,nz,nx)
    end if



  ! Apply taper to image 
  allocate(imgtaper(20))
  do i=1,20  
   imgtaper(i) = 0.05*(20-i)
  end do
  do iz=1,nz
     do i=1,20
        image(iz,nx-npad+i) = image(iz,nx-npad)*imgtaper(i)
        image(iz,npad-i+1) = image(iz,npad+1)*imgtaper(i)
     end do
  end do
  image(1:npad+4,:) = 0.
  do iz=1,20
     image(npad+4+iz,:) = image(npad+4+iz,:)*imgtaper(20-i)
  end do


  ! Initiate propagation
  call init_propagate(vel,oz,ox,dz,dx,dt_data,nt_data,wfldmovie,souWfile_,recWfile_,imgWfile_,vmax,npad) 

  ! Initiate source parameters
  call source_init(wavelet,xsrc,zsrc,tdelay)
  
  ! Initiate receiver parameters
  call receiver_init(.true.,zsrc,xsrc,orx,drx,dat)

  ! Migrate a shot
  call born_model(image,dat)


  ! Datamodel
   if (DW) then
      allocate(DWarray(nt_data,nrx))
      call sep_read(DWarray(1:nt_data,1:nrx),"DWfile")
      do ix=1,nrx
         do it=1,nt_data
            dat(it,ix) = dat(it,ix)*DWarray(it,ix)
         end do
      end do
   end if

  ! Output data
    call auxputch("n1",'i',nt_data,outfile_)
    call auxputch("n2",'i',nrx,outfile_)
    call auxputch("d1",'f',dt_data,outfile_)
    call auxputch("d2",'f',drx,outfile_)
    call auxputch("o1",'f',ot_data,outfile_)
    call auxputch("o2",'f',orx,outfile_)
    call auxputch("label1",'s','t(s)',outfile_) 
    call auxputch("label2",'s','rx(m)',outfile_) 
  call sep_write(dat,outfile_)

 ! call sleep(600)

end program bornmodel                                         
