!RTM code
!Actual migration by cross correlating the source and the receiver wavefields
!Initial version written by Noha in September 2012
!Modified by Mandy in November 2012

Program migration

  use sep
  use prop_mod
  implicit none

  logical(4) :: success
  logical                               :: wfldmovie, VDer, DW
  real                                  :: dx, dz, ox, oz, ost, dst, orx, drx, ot_data, dt_data
  real                                  :: xsrc, zsrc, vmax,tdelay
  integer                               :: nnx, nnz, nx, nz, nst, nt_data, nrx, nop, npad
  integer            			:: ix, iz , it
  real, allocatable, dimension(:,:)     :: vel, mwarr, dat, DWarray
  real, allocatable, dimension(:,:)     :: image
  real, allocatable, dimension(:)       ::wavelet
  character(len=128)                    :: souWfile_, imgWfile_, recWfile_
  character(len=128)                    :: outfile_

  call sep_init()

  ! read in velocity model
  call from_history("n1", nz)
  call from_history("o1", oz)
  call from_history("d1", dz)
  call from_history("n2", nx)
  call from_history("o2", ox)
  call from_history("d2", dx)

  ! read source wavelet information
  call from_aux("aux", "n1", nst)
  call from_aux("aux", "o1", ost)
  call from_aux("aux", "d1", dst)

  ! read shot gather (data) file info
  call from_aux("datafile", "n1", nt_data)
  call from_aux("datafile", "o1", ot_data)
  call from_aux("datafile", "d1", dt_data)
  call from_aux("datafile", "n2", nrx)
  call from_aux("datafile", "o2", orx)    ! orx is defined to be dist from xsrc
  call from_aux("datafile", "d2", drx)

  write(0,*) 'nt_data, ot, dt', nt_data, ot_data, dt_data
  write(0,*) 'nrx, orx, drx', nrx, orx, drx
  

  ! read source location
  call from_param("zsrc",zsrc)
  call from_param("xsrc",xsrc)
  call from_param("npad",npad,30)
  
  ! Read wavefield movie switch
  call from_param("wfldout",wfldmovie,.false.)
  call from_param("VDer",VDer,.false.)
  call from_param("DW",DW,.false.)
  call from_param("souwfld",souWfile_,"souwfld.H")
  call from_param("imgwfld",imgWfile_,"imgwfld.H")  
  call from_param("recwfld",recWfile_,"recwfld.H")
  call from_param("outfile",outfile_)
  call from_param("tdelay",tdelay)



  if (dst .ne. dt_data) call erexit('dst should equal to dt_data')

  nnz=nz
  nnx=nx
  nz = nnz+2*npad
  nx = nnx+2*npad



  ! Allocate memory
  allocate(vel(nz,nx),dat(nt_data,nrx),image(nz,nx),wavelet(nt_data))


  ! Read velocity model
  call sep_read(vel(npad+1:nz-npad,npad+1:nx-npad));  call from_param("vmax",vmax);
  
  ! Read source wavelet
  call sep_read(wavelet,"aux")
  
  ! Read data
  call sep_read(dat,"datafile")

  ! Datamodel
   if (DW) then
      allocate(DWarray(nt_data,nrx))
      call sep_read(DWarray(1:nt_data,1:nrx),"DWfile")
      do ix=1,nrx
         do it=1,nt_data
            dat(it,ix) = dat(it,ix)*DWarray(it,ix)
         end do
      end do
   end if

  ! Initiate propagation
  call init_propagate(vel,oz,ox,dz,dx,dt_data,nt_data,wfldmovie,souWfile_,recWfile_,imgWfile_,vmax,npad) 

  ! Initiate source parameters
  call source_init(wavelet,xsrc,zsrc,tdelay)
  
  ! Initiate receiver parameters
  call receiver_init(.true.,zsrc,xsrc,orx,drx,dat)

  ! Migrate a shot
  call migrate_shot(image)

  write(0,*) 'pass 1'

!   ! Apply filter to model
    if (VDer) then
       call Ver2ndDer(.true.,image,nz,nx)
    end if

  write(0,*) 'pass 2'

  ! Apply model weighting
  allocate(mwarr(nnz,nnx))
  mwarr=1.
  call sep_read(mwarr(1:nnz,1:nnx),"mwfile")
  do ix=npad+1,nx-npad
     do iz=npad+1,nz-npad
        image(iz,ix) = image(iz,ix)*mwarr(iz-npad,ix-npad)
     end do
  end do

  write(0,*) 'pass 3', outfile_


  ! output image
    call writeimageheader(outfile_,nz-2*npad,nx-2*npad)
  call sep_write(image(1+npad:nz-npad,1+npad:nx-npad),outfile_)

end program                                
