module waveprop3d_mod

  use type3d_mod
  use omp_lib

  implicit none

      real                    :: W0_2d12, W1_2d12, W2_2d12, W3_2d12
      real                    :: W4_2d12, W5_2d12, W6_2d12, FN_2d12
      real                    :: W0_2d16, W1_2d16, W2_2d16, W3_2d16, W4_2d16
      real                    :: W5_2d16, W6_2d16, W7_2d16, W8_2d16

!     ------------------------------------------------------
!     16th order optimized FD weights (2nd):
!     ------------------------------------------------------
      PARAMETER  ( W0_2D16 = -3.22511718414079 )  
      PARAMETER  ( W1_2D16 =  1.93606382829618 )  
      PARAMETER  ( W2_2D16 = -0.43844416993301 )
      PARAMETER  ( W3_2D16 =  0.16442503430510 )
      PARAMETER  ( W4_2D16 = -0.07205642556714 )
      PARAMETER  ( W5_2D16 =  0.03264972066062 )
      PARAMETER  ( W6_2D16 = -0.01416701049701 )
      PARAMETER  ( W7_2D16 =  0.00516908814658 )
      PARAMETER  ( W8_2D16 = -0.00108147334093 )
!      PARAMETER  ( FN_2D16 = 1.0 / 1.46229534342707 )
  
!
!     ------------------------------------------------------
!     12th order optimized FD weights, extreme (2nd):
!     ------------------------------------------------------
      PARAMETER  ( W0_2D12  = -3.16513614080052 )  
      PARAMETER  ( W1_2D12  =  1.87834860513654 )  
      PARAMETER  ( W2_2D12  = -0.38710535345218 )
      PARAMETER  ( W3_2D12  =  0.12241455362069 )
      PARAMETER  ( W4_2D12  = -0.04073045512005 )
      PARAMETER  ( W5_2D12  =  0.01159369937042 )
      PARAMETER  ( W6_2D12  = -0.00195297915516 )
      PARAMETER  ( FN_2D12  = 1.0 / 1.41857564413310 )



contains

!==========================================================

  subroutine Gaussian3D(adj,m,nz,nx,ny)
  ! Take a 3D Gaussian and its adjoint

  logical                   :: adj
  integer                   :: nz, nx, ny, ix, iy, iz, nop,wd, i, j, o, k, l, p
  real                      :: dx, dz, width2, val, val2, val3, norm, a, kk 
  real, dimension(:,:,:), allocatable :: gauss
  real, dimension(:,:,:)      :: m
  real, dimension(:,:,:), allocatable  :: d


  wd = 2
  norm = 0.0
  width2 = 0.0
  val = 0.0

  write(0,*) 'Inside Gaussian3D'

  allocate(d(nz,nx,ny))
  allocate(gauss(wd+1,wd+1,wd+1))  

  do iy=1,ny
  do ix=1,nx
     do iz=1,nz
        d(iz,ix,iy)=0.
     end do
  end do
  end do
 
  write(0,*) 'Constructing filters'
  ! filter is only 1/8th of the entire coverage because it is symmetric about zero
  ! the center of the filter is gauss(1,1,1) (largest amp), and the smallest value is gauss(wd+1,wd+1,wd+1)
  a = 5.
  width2 = 2.0*real(wd+1)**2
  kk = a*a/width2

  do o=1,wd+1
  do j=1,wd+1
    do i=1,wd+1
       val = real(i-1)
       val2 = real(j-1)
       val3 = real(o-1)
       gauss(i,j,o) = exp(-kk*val*val-kk*val2*val2-kk*val3*val3)
       norm = norm + gauss(i,j,o)
    end do
  end do
  end do

  gauss = gauss/norm/8.0
  

  if (1 == 1) then
     write(0,*) 'gauss filter'
    do o=1,wd+1
    do j=1,wd+1
    do i=1,wd+1
       write(0,*) 'i, j, o', i, j, o, gauss(i,j,o)
    end do
    end do
    end do
    write(0,*) 'sum Gauss', sum(gauss)
  end if


  write(0,*) 'Applying filter'

  if (.not. adj) then
  do iy=1+wd, ny-wd  
  do ix=1+wd, nx-wd  
     do iz=1+wd, nz-wd
       do p=0,wd
       do k=0,wd
         do l=0,wd
          d(iz,ix,iy) = d(iz,ix,iy) + gauss(k,l,p)*(m(iz+k,ix+l,iy+p) + m(iz-k,ix+l,iy+p) + & 
              m(iz+k,ix-l,iy+p) + m(iz-k,ix-l,iy+p)+ m(iz+k,ix+l,iy-p) + m(iz-k,ix+l,iy-p) + & 
              m(iz+k,ix-l,iy-p) + m(iz-k,ix-l,iy-p) )      
         end do
       end do
       end do
     end do
   end do
   end do


  else

  do iy=1+wd, ny-wd  
  do ix=1+wd, nx-wd  
     do iz=1+wd, nz-wd
       do p=0,wd
       do k=0,wd
         do l=0,wd
           d(iz+k,ix+l,iy+p) = d(iz+k,ix+l,iy+p) + gauss(k,l,p)*m(iz,ix,iy)
           d(iz-k,ix+l,iy+p) = d(iz-k,ix+l,iy+p) + gauss(k,l,p)*m(iz,ix,iy)
           d(iz+k,ix-l,iy+p) = d(iz+k,ix-l,iy+p) + gauss(k,l,p)*m(iz,ix,iy)
           d(iz-k,ix-l,iy+p) = d(iz-k,ix-l,iy+p) + gauss(k,l,p)*m(iz,ix,iy)     

           d(iz+k,ix+l,iy-p) = d(iz+k,ix+l,iy-p) + gauss(k,l,p)*m(iz,ix,iy)
           d(iz-k,ix+l,iy-p) = d(iz-k,ix+l,iy-p) + gauss(k,l,p)*m(iz,ix,iy)
           d(iz+k,ix-l,iy-p) = d(iz+k,ix-l,iy-p) + gauss(k,l,p)*m(iz,ix,iy)
           d(iz-k,ix-l,iy-p) = d(iz-k,ix-l,iy-p) + gauss(k,l,p)*m(iz,ix,iy) 
         end do
       end do
       end do
     end do
   end do
   end do
  end if

    write(0,*) 'm, d', m(iz/6,ix/2,iy/2), d(iz/6,ix/2,iy/2)


   do iy=1,ny
    do ix=1,nx
     do iz=1,nz
        m(iz,ix,iy) = m(iz,ix,iy) - d(iz,ix,iy)
     end do
    end do
   end do



  write(0,*) 'outputting Gaussian', sum(m)

  end subroutine


!==========================================================


  subroutine Laplacian3D(adj,m,nz,nx,ny)
  ! Take a 3D laplacian and its adjoint

  logical                   :: adj
  integer                   :: nz, nx, ny, ix, iy, iz, nop
  real, dimension(:,:,:)      :: m
  real, dimension(:,:,:), allocatable  :: d

  write(0,*) 'Inside Laplacian3D'

  allocate(d(nz,nx,ny))
  
  do iy=1,ny
    do ix=1,nx
      do iz=1,nz
        d(iz,ix,iy)=0.
      end do
    end do
  end do
 
  nop=1 

  if (.not. adj) then
    do iy=1+nop,ny-nop
      do ix=1+nop,nx-nop
         do iz=1+nop, nz-nop

          d(iz,ix,iy) = d(iz,ix,iy) + 6.*m(iz,ix,iy) - m(iz+1,ix,iy) - m(iz-1,ix,iy) - &
                     m(iz,ix+1,iy) - m(iz,ix-1,iy) - m(iz,ix,iy+1) - m(iz,ix,iy-1)  

       enddo
      enddo
   enddo


  else

   do iy=1+nop,ny-nop
    do ix=1+nop,nx-nop
       do iz=1+nop, nz-nop
          d(iz,ix,iy) = d(iz,ix,iy) + 6.*m(iz,ix,iy)
          d(iz+1,ix,iy) = d(iz+1,ix,iy) + -1.*m(iz,ix,iy)
	  d(iz-1,ix,iy) = d(iz-1,ix,iy) + -1.*m(iz,ix,iy)
          d(iz,ix+1,iy) = d(iz,ix+1,iy) + -1.*m(iz,ix,iy)
          d(iz,ix-1,iy) = d(iz,ix-1,iy) + -1.*m(iz,ix,iy)
          d(iz,ix,iy+1) = d(iz,ix,iy-1) + -1.*m(iz,ix,iy)
          d(iz,ix,iy-1) = d(iz,ix,iy+1) + -1.*m(iz,ix,iy)
       enddo
    enddo
   end do


  end if

   do iy=1,ny
    do ix=1,nx
     do iz=1,nz
        m(iz,ix,iy)=d(iz,ix,iy)
     end do
    end do
   end do

  write(0,*) 'Laplace 3D', adj

  end subroutine

!==========================================================

  subroutine Ver2ndDer3D(adj,m,nz,nx,ny)
  ! Take the second derivative vertically and its adjoint

  logical                   :: adj
  integer                   :: nz, nx, ny, ix, iy, iz 
  real, dimension(:), allocatable :: temp
  real, dimension(:,:,:)      :: m

  write(0,*) 'Inside Ver2ndDer3D'

  allocate(temp(nz))

  do iy=1,ny
    do ix=1,nx
      temp(:) = m(:,ix,iy)  
      m(:,ix,iy) = 0.
      do iz=2,nz-1
         if (adj) then
            m(iz,ix,iy) = m(iz,ix,iy) + (-2.*temp(iz) + temp(iz+1) + temp(iz-1))
         else
          m(iz,ix,iy) = m(iz,ix,iy) + -2*temp(iz)
          m(iz,ix,iy) = m(iz,ix,iy) + temp(iz+1)
          m(iz,ix,iy) = m(iz,ix,iy) + temp(iz-1)
         end if
      end do
    end do
  end do


  end subroutine 

!==========================================================

  subroutine Ver2ndDer1D(adj,m,nt)
  ! Take the second derivative vertically and its adjoint

  logical                   :: adj
  integer                   :: nt, it 
  real, dimension(:), allocatable :: temp
  real, dimension(:)      :: m

  write(0,*) 'Inside Ver2ndDer1D'

  allocate(temp(nt))

      temp(:) = m(:)  
      m(:) = 0.
      do it=2,nt-1
         if (adj) then
            m(it) = m(it) + (-2.*temp(it) + temp(it+1) + temp(it-1))
         else
          m(it) = m(it) + -2*temp(it)
          m(it) = m(it) + temp(it+1)
          m(it) = m(it) + temp(it-1)
         end if
      end do

    deallocate(temp)

  end subroutine 


!==========================================================

   subroutine getfilename(base_,swout_,ifile)

        integer      :: ifile,p1,d2
        character(len=128)   :: base_,swout_
        character*3          :: aa

        p1 = scan(base_,".H")

        if (ifile .eq. 0) then
          swout_(:p1+1) = base_(:p1+1)
        else
  	  swout_(:p1-1) = base_(:p1-1)
          swout_(p1:) = " "
          write(aa,1000) ifile
          1000 format(i3)

          d2=1
          if (ifile .gt. 9) d2=2
          if (ifile .gt. 99) d2=3

          swout_(p1:p1) ="["
          swout_(p1+1:p1+d2)=aa(3-d2+1:3)
          swout_(p1+d2+1:p1+d2+1)="]"
          swout_(p1+d2+2:) = ".H"
        end if

   end subroutine

!=========================================
    subroutine DPdim1(a,b,n1,tot)
   ! sum the element of a*b
   ! where a*b is element-wise sum
	
	integer        :: n1,i1
        real, dimension(:) :: a, b
        real*8         :: tot, val
        real*8         :: tmpP, errP, totP
        real*8         :: tmpM, errM, totM

	tot=0;  val=0; 
        totP=0.; tmpP=0.; errP=0.;
        totM=0.; tmpM=0.; errM=0.;

        do i1=1,n1
	   val = a(i1)*b(i1)
	   if (val .gt. 0.00000) then
              tmpP=totP+val+errP
              errP=(totP-tmpP)+val
              totP=tmpP
           else 
              tmpM=totM+val+errM
              errM=(totM-tmpM)+val
              totM=tmpM
           end if
        end do
    
	tot = totP+totM

    end subroutine
!==========================================================
   subroutine DPdim2(a,b,n1,n2,tot)
	
	integer        :: n1,n2,i1,i2
        real, dimension(:,:) :: a, b
        real*8         :: tmp, err, tot
        
          do i2=1,n2
            do i1=1,n1
               tmp=tot+a(i1,i2)*b(i1,i2)+err
               err=(tot-tmp)+a(i1,i2)*b(i1,i2)
               tot=tmp
            end do
          end do
    
    end subroutine
!==========================================================
   subroutine DPdim3(a,b,n1,n2,n3,tot)
   ! sum the element of a*b
   ! where a*b is element-wise sum
	
	integer        :: n1,n2,n3,i1,i2,i3
        real, dimension(:,:,:) :: a, b
        real*8         :: tot, val
        real*8         :: tmpP, errP, totP
        real*8         :: tmpM, errM, totM

	tot=0;  val=0; 
        totP=0.; tmpP=0.; errP=0.;
        totM=0.; tmpM=0.; errM=0.;

        do i3=1,n3
          do i2=1,n2
            do i1=1,n1
	       val = a(i1,i2,i3)*b(i1,i2,i3)
	       if (val .gt. 0.00000) then
                  tmpP=totP+val+errP
                  errP=(totP-tmpP)+val
                  totP=tmpP
               else 
                  tmpM=totM+val+errM
                  errM=(totM-tmpM)+val
                  totM=tmpM
               end if
            end do
          end do
        end do
    
	tot = totP+totM

    end subroutine
!==========================================================
    subroutine calcBksz(n, nblocks, grid_size,block_size,nnop)
	! Calculate block size 
        
       integer          :: n, nblocks, grid_size, block_size, nnop

       if (mod(n,nblocks) .ne. 0) then
          write(0,*) 'size n=', n, 'nblocks=', nblocks
          call erexit('Error: make sure nblocks divide n')

       end if

  
       grid_size=(n)/nblocks
       block_size = grid_size
 
       if (1.5*nnop .gt. grid_size) call erexit('Error: block size < 1.5*nop')
       if (grid_size .lt. 1) call erexit("Error: model size is too small")

    end subroutine


!==========================================================
    subroutine calcBkNum(nnop,xx)
	! Calculate number of blocks based on block size and grid_size
        

       type(blcIndex)  :: xx
       integer          :: nnop

  
       xx%effn = xx%n - 2*nnop
       xx%effbsz = xx%bsz -2*nnop
       xx%nblocks = ceiling(real(xx%effn)/real(xx%effbsz)) 
       xx%effbszlast = mod(xx%effn,xx%effbsz)
       if (xx%effbszlast .eq. 0) xx%effbszlast = xx%effbsz
       xx%bszlast = xx%effbszlast+2*nnop
  
!       write(0,*) 'n, bsz, nop', xx%n, xx%bsz, nnop
!       write(0,*) 'effn,effbsz,nblocks', xx%effn, xx%effbsz, xx%nblocks
 !      write(0,*) 'effbszlast, bszlast', xx%effbszlast, xx%bszlast

       if (xx%effbsz .le. 5) call erexit('calcBkNum error: effbsz is too small')
       if (xx%effn .lt. 1) call erexit('calcBkNum error: effn is too small')

       
 

    end subroutine

!==========================================================
    subroutine calcBkNum2(effn,effbsz,nblocks,effbszlast,bszlast,n,bsz,nnop)
        ! This is the same as calcBkNum but without the use of type variable
        

       integer          :: nnop, effn, effbsz, nblocks
       integer          :: effbszlast, bszlast,n,bsz

  
       effn = n - 2*nnop
       effbsz = bsz -2*nnop
       nblocks = ceiling(real(effn)/real(effbsz)) 
       effbszlast = mod(effn,effbsz)
       if (effbszlast .eq. 0) effbszlast = effbsz
       bszlast = effbszlast+2*nnop
  
!       write(0,*) 'n, bsz, nop', n, bsz, nnop
!       write(0,*) 'effn,effbsz,nblocks', effn, effbsz, nblocks
 !      write(0,*) 'effbszlast, bszlast', effbszlast, bszlast

       if (effbsz .le. 5) call erexit('calcBkNum error: effbsz is too small')
       if (effn .lt. 1) call erexit('calcBkNum error: effn is too small')

       
 

    end subroutine
!==========================================================
    subroutine calcIndex(x,iblock, nnop)

       type(blcIndex)   :: x
       integer          :: iblock, nnop

       x%iib = (iblock-1)*x%grid_size
       x%padb = min((iblock-1)*2*nnop,nnop)
       x%pade = min((x%nblocks-iblock)*2*nnop,nnop)
       x%ib = x%iib+1-x%padb
       x%ie = x%iib+x%block_size+x%pade
       x%dbloc=x%block_size+x%pade+x%padb-2*nnop


    end subroutine

!==========================================================
    subroutine calcIndex2(x,iblock, nnop)
    ! calculate ib and ie for the block
    ! This version input as fixed block size and then sub-divide the domain
    ! it is called after calcBlkNum function

       type(blcIndex)   :: x
       integer          :: iblock, nnop, curreffbsz

       if (iblock .eq. x%nblocks) then
          x%ib = 1+nnop+(x%nblocks-1)*x%effbsz
          x%ie = x%ib+x%effbszlast-1
          x%dbloc = x%effbszlast
       else
          x%ib = 1+(iblock-1)*x%effbsz+nnop
          x%ie = x%ib + x%effbsz -1
          x%dbloc = x%effbsz 
       end if
       

    end subroutine
!==========================================================
    subroutine calcIndex3(nblocks,ib,ie,effbsz,effbszlast,dbloc,iblock, nnop)
    ! This is the same as calcIndex2, but do not use the type structure
    ! calculate ib and ie for the block
    ! This version input as fixed block size and then sub-divide the domain
    ! it is called after calcBlkNum function

       type(blcIndex)   :: x
       integer          :: iblock, nnop, curreffbsz
       integer          :: nblocks,ib,ie,effbsz,effbszlast,dbloc

       if (iblock .eq. nblocks) then
          ib = 1+nnop+(nblocks-1)*effbsz
          ie = ib+effbszlast-1
          dbloc = effbszlast
       else
          ib = 1+(iblock-1)*effbsz+nnop
          ie = ib + effbsz -1
          dbloc = effbsz 
       end if
       

    end subroutine
!==========================================================
!! Taper Setup Cosine

  subroutine tapersetupCos3d(bndz,bndx,bndy,ztaper,xtaper,ytaper,cent, A)

    integer                     :: iz, ix, iy, bndz, bndx, bndy
    real, dimension(:)          :: xtaper,ytaper,ztaper
    real                       :: dmax, dist, cent, A
    real                        :: pi=3.14159265359


     dmax = max(bndx, bndz)
!    dmax = sqrt(pi*bndx*bndx + bndz*bndx)
    


       ! Left and right edge
       do iz=1,bndz
          ztaper(iz) = cent + A*cos(pi*(bndz-iz)/bndz)
       end do

       ! top and bottom edge
       do ix=1,bndx
          xtaper(ix) = cent + A*cos(pi*(bndx-ix)/bndx)
       end do

       ! front and back edge
       do iy=1,bndy
          ytaper(iy) = cent + A*cos(pi*(bndy-iy)/bndy)
       end do


 

  end subroutine

!============================================================
! Tapering
! Version 2 has reflective top boundary
! no padding is applied to the top boundary
! nz = nnz + bndz + nop

 subroutine taperingCos3dV2(nz,nx,ny,nop, bndz,bndx,bndy,ztaper,xtaper,ytaper,uo)

    integer  :: iz, ix, iy, nz,nx, ny,bndz,bndx, bndy,nop
    real, dimension(:) :: xtaper,ztaper, ytaper
    real, dimension(:,:,:) :: uo


       do iy=1,ny
         do iz=1+nop,nz
          uo(iz, 1:bndx,iy) = uo(iz, 1:bndx,iy) * xtaper(1:bndx:+1);              !left
	  uo(iz,nx-bndx+1:nx,iy) = uo(iz,nx-bndx+1:nx,iy) * xtaper(bndx:1:-1)     !right
         end do
       end do

       do ix=1,nx
         do iz=1+nop,nz
          uo(iz, ix, 1:bndy) = uo(iz, ix,1:bndy) * ytaper(1:bndy:+1);              !front
	  uo(iz, ix, ny-bndy+1:ny) = uo(iz, ix,ny-bndy+1:ny) * ytaper(bndy:1:-1)     !back
         end do
       end do


       do iy=1,ny
         do ix=1,nx
            uo(nz-bndz+1:nz,ix,iy) = uo(nz-bndz+1:nz,ix,iy) * ztaper(bndz:1:-1)  ! bottom
         end do
       end do


  end subroutine

!============================================================
! Tapering
! Version 3 is an alternate version (I am a bit confused in how the tapering should work in 3d)
! bndz padding is applied to the top boundary
! nz = nnz + 2*bndz 
! nx = nnx + 2*bndx
! ny = nny + 2*bndy


 subroutine taperingCos3d(nz,nx,ny,bndz,bndx,bndy,ztaper,xtaper,ytaper,uo)

    integer  :: iz, ix, iy, nz,nx, ny,bndz,bndx, bndy,nop
    real, dimension(:) :: xtaper,ztaper, ytaper
    real, dimension(:,:,:) :: uo

       do iy=1,ny
         do iz=1,nz
          uo(iz, 1:bndx,iy) = uo(iz, 1:bndx,iy) * xtaper(1:bndx:+1);              !left
	  uo(iz,nx-bndx+1:nx,iy) = uo(iz,nx-bndx+1:nx,iy) * xtaper(bndx:1:-1)     !right
         end do
       end do

       do ix=1,nx
         do iz=1,nz
          uo(iz, ix, 1:bndy) = uo(iz, ix,1:bndy) * ytaper(1:bndy:+1);              !front
	  uo(iz, ix, ny-bndy+1:ny) = uo(iz, ix,ny-bndy+1:ny) * ytaper(bndy:1:-1)     !back
         end do
       end do


       do iy=1,ny
         do ix=1,nx
            uo(1:bndz,ix,iy) = uo(1:bndz,ix,iy) * ztaper(1:bndz:+1);             ! top
            uo(nz-bndz+1:nz,ix,iy) = uo(nz-bndz+1:nz,ix,iy) * ztaper(bndz:1:-1)  ! bottom
         end do
       end do

  end subroutine
!===========================================================
!! take 1st derivative

  subroutine FirstDer(adj,souIn,souOut,nt,dt)

    logical :: adj
    integer :: i, nt
    real    :: dt, rdt
    real, dimension(:) :: souIn, souOut
    real, dimension(:), allocatable :: temp

   allocate(temp(nt))
   temp(1:nt) = souIn(1:nt)
    
    souOut(1:nt) = 0.
    
    
    if (.not. adj) then
       
      do i=1,nt-1
         souOut(i+1) = souOut(i+1)+ temp(i)
         souOut(i)   = souOut(i) - temp(i)
      end do
      
    else
       
      do i=1,nt-1
         souOut(i+1) = souOut(i+1) + temp(i+1)-temp(i)
      end do
         
    end if    

    deallocate(temp)


  end subroutine 
!===========================================================
!!--------------------------------------------------------------------------
!! 16th Order approximation to the 3D Laplacian
!! Coefficients are optimized to minimize dispersion
!! This code uses domain decomposition with secified block size
!! Expanding domain in z


  subroutine laplac3d16(adj, uo,up, vp2, nx, ny,nz, nzmax, xbsz,ybsz,zbsz, idx,idy,idz,nop, iit,nodes)

    logical    :: adj
    integer    :: nop, iit, inode, nodes
    integer    :: iblockx, ix, jx, nx
    integer    :: iblocky, iy, jy, ny
    integer    :: iblockz, iz, jz, nz, nzmax
    real, dimension(:,:,:) :: uo,up, vp2
    real       :: c0, idx, idy, idz, v2uo, v2up
    real       :: c1x, c2x, c3x, c4x, c5x, c6x, c7x, c8x
    real       :: c1y, c2y, c3y, c4y, c5y, c6y, c7y, c8y
    real       :: c1z, c2z, c3z, c4z, c5z, c6z, c7z, c8z


    integer    :: xdbloc, xib, xie, nbkx, xeffbszlast, xeffbsz
    integer    :: ydbloc, yib, yie, nbky, yeffbszlast, yeffbsz
    integer    :: zdbloc, zib, zie, nbkz, zeffbszlast, zeffbsz
    integer    :: xeffn, xbszlast, xbsz
    integer    :: yeffn, ybszlast, ybsz
    integer    :: zeffn, zbszlast, zbsz

    real, dimension(:,:,:), allocatable   :: Buo, Bup

    ! Define cofficients 
  c0 = W0_2d16*(idx+idy+idz)
  c1x = W1_2d16*idx;   c1y = W1_2d16*idy;   c1z =W1_2d16*idz; 
  c2x = W2_2d16*idx;   c2y = W2_2d16*idy;   c2z = W2_2d16*idz;  
  c3x = W3_2d16*idx;   c3y = W3_2d16*idy;   c3z = W3_2d16*idz;
  c4x = W4_2d16*idx;   c4y = W4_2d16*idy;   c4z = W4_2d16*idz; 
  c5x = W5_2d16*idx;   c5y = W5_2d16*idy;   c5z = W5_2d16*idz; 
  c6x = W6_2d16*idx;   c6y = W6_2d16*idy;   c6z = W6_2d16*idz; 
  c7x = W7_2d16*idx;   c7y = W7_2d16*idy;   c7z = W7_2d16*idz; 
  c8x = W8_2d16*idx;   c8y = W8_2d16*idy;   c8z = W8_2d16*idz; 

  ! Calculate Number of Blocks
  call calcBkNum2(xeffn,xeffbsz,nbkx,xeffbszlast,xbszlast,nx,xbsz,nop)
  call calcBkNum2(yeffn,yeffbsz,nbky,yeffbszlast,ybszlast,ny,ybsz,nop)
  call calcBkNum2(zeffn,zeffbsz,nbkz,zeffbszlast,zbszlast,nzmax,zbsz,nop)

  ! Initialize , zero the output first
  do iy=1,ny
    do ix=1,nx
      do iz=1,nz
         up(iz,ix,iy) = 0.
      end do
    end do
  end do



!  write(0,*) 'nbkz,nbkx,nbky',nbkz,nbkx,nbky
!  write(0,*) 'zbsz,xbsz,ybsz',zbsz,xbsz,ybsz

    if (adj) then    !! RTM

!        write(0,*) 'in the Adjoint direction', iit, nodes
!        write(0,*) 'yy',ybsz,ydbloc

	if (allocated(Buo)) deallocate(Buo); allocate(Buo(zbsz,xbsz,ybsz)) 

	!$OMP PARALLEL DO DEFAULT(shared) schedule(dynamic,1) num_threads(nodes) PRIVATE(inode, iblocky,iblockx,iblockz,xib, yib,zib,xie,yie,zie,xdbloc,ydbloc,zdbloc,iz, jz,ix,jx,iy,jy,Buo) 

        do iblocky=1,nbky
           call calcIndex3(nbky,yib,yie,yeffbsz,yeffbszlast,ydbloc,iblocky,nop)
!	   inode = omp_get_thread_num()
!	   write(0,*) 'block #:', iblocky, 'out of', nbky, 'thread:', inode

	  do iblockx=1,nbkx
 	     call calcIndex3(nbkx,xib,xie,xeffbsz,xeffbszlast,xdbloc,iblockx,nop)

            do iblockz=1,nbkz
	       call calcIndex3(nbkz,zib,zie,zeffbsz,zeffbszlast,zdbloc,iblockz,nop)

	       Buo(1:zdbloc+2*nop,1:xdbloc+2*nop,1:ydbloc+2*nop) =  &
                             uo(zib-nop:zie+nop,xib-nop:xie+nop,yib-nop:yie+nop)


               do iy = 1+nop,ydbloc+nop
                  jy = iy+yib-1-nop
                 do ix = 1+nop,xdbloc+nop
                    jx = ix+xib-1-nop
                   do iz = 1+nop,zdbloc+nop
                    jz = iz+zib-1-nop


                   up(jz,jx,jy) = 2*Buo(iz, ix, iy) + &
                                 vp2(jz, jx, jy)*(c0* Buo(iz,ix,iy) +&
                   c1z*(Buo(iz-1,ix  ,iy  ) + Buo(iz+1,ix  ,iy  )) + &
                   c1x*(Buo(iz  ,ix-1,iy  ) + Buo(iz  ,ix+1,iy  )) + &
                   c1y*(Buo(iz  ,ix  ,iy-1) + Buo(iz  ,ix  ,iy+1)) + &
                   c2z*(Buo(iz-2,ix  ,iy  ) + Buo(iz+2,ix  ,iy  )) + &
                   c2x*(Buo(iz  ,ix-2,iy  ) + Buo(iz  ,ix+2,iy  )) + &
                   c2y*(Buo(iz  ,ix  ,iy-2) + Buo(iz  ,ix  ,iy+2)) + &
                   c3z*(Buo(iz-3,ix  ,iy  ) + Buo(iz+3,ix  ,iy  )) + &
                   c3x*(Buo(iz  ,ix-3,iy  ) + Buo(iz  ,ix+3,iy  )) + &
                   c3y*(Buo(iz  ,ix  ,iy-3) + Buo(iz  ,ix  ,iy+3)) + &
                   c4z*(Buo(iz-4,ix  ,iy  ) + Buo(iz+4,ix  ,iy  )) + &
                   c4x*(Buo(iz  ,ix-4,iy  ) + Buo(iz  ,ix+4,iy  )) + &
                   c4y*(Buo(iz  ,ix  ,iy-4) + Buo(iz  ,ix  ,iy+4)) + & 
                   c5z*(Buo(iz-5,ix  ,iy  ) + Buo(iz+5,ix  ,iy  )) + &
                   c5x*(Buo(iz  ,ix-5,iy  ) + Buo(iz  ,ix+5,iy  )) + &
                   c5y*(Buo(iz  ,ix  ,iy-5) + Buo(iz  ,ix  ,iy+5)) + & 
                   c6z*(Buo(iz-6,ix  ,iy  ) + Buo(iz+6,ix  ,iy  )) + &
                   c6x*(Buo(iz  ,ix-6,iy  ) + Buo(iz  ,ix+6,iy  )) + &
                   c6y*(Buo(iz  ,ix  ,iy-6) + Buo(iz  ,ix  ,iy+6)) + & 
                   c7z*(Buo(iz-7,ix  ,iy  ) + Buo(iz+7,ix  ,iy  )) + &
                   c7x*(Buo(iz  ,ix-7,iy  ) + Buo(iz  ,ix+7,iy  )) + &
                   c7y*(Buo(iz  ,ix  ,iy-7) + Buo(iz  ,ix  ,iy+7)) + & 
                   c8z*(Buo(iz-8,ix  ,iy  ) + Buo(iz+8,ix  ,iy  )) + &
                   c8x*(Buo(iz  ,ix-8,iy  ) + Buo(iz  ,ix+8,iy  )) + &
                   c8y*(Buo(iz  ,ix  ,iy-8) + Buo(iz  ,ix  ,iy+8)))


                  end do
                end do
              end do

	    end do
	   end do         
	 end do
         !$OMP END PARALLEL DO

    else  !! FORWARD

   !  write(0,*) 'sum uo', sum(uo), sum(vp2)


       if (allocated(Bup)) deallocate(Bup); allocate(Bup(zbsz,xbsz,ybsz)) 


       !$OMP PARALLEL DO DEFAULT(shared) schedule(dynamic,1) num_threads(nodes) PRIVATE(inode, iblocky,iblockx,iblockz,xib, yib,zib,xie,yie,zie,xdbloc,ydbloc,zdbloc,iz, jz,ix,jx,iy,jy,Bup,v2uo) 

        do iblocky=1,nbky
           call calcIndex3(nbky,yib,yie,yeffbsz,yeffbszlast,ydbloc,iblocky,nop)

	  do iblockx=1,nbkx
 	     call calcIndex3(nbkx,xib,xie,xeffbsz,xeffbszlast,xdbloc,iblockx,nop)

            do iblockz=1,nbkz
	       call calcIndex3(nbkz,zib,zie,zeffbsz,zeffbszlast,zdbloc,iblockz,nop)
 
                ! zeros the block first
                do iy=1,ybsz
                  do ix=1,xbsz
                    do iz=1,zbsz
              	      Bup(iz,ix,iy) = 0.
                    end do
                  end do
                end do


               do iy = 1+nop,ydbloc+nop
                  jy = iy+yib-1-nop
                 do ix = 1+nop,xdbloc+nop
                    jx = ix+xib-1-nop
                   do iz = 1+nop,zdbloc+nop
                    jz = iz+zib-1-nop

                   v2uo = vp2(jz,jx,jy)*uo(jz,jx,jy)

!                   if ((iy .eq. 20) .and. (ix .eq. 20) .and. (iz .eq. 20)) then
!                      write(0,*) v2uo, vp2(jz,jx,jy), uo(jz,jx,jy)
!                   end if



                   Bup(iz,ix,iy) = Bup(iz,ix,iy) + 2*uo(jz,jx,jy)
                   Bup(iz,ix,iy) = Bup(iz,ix,iy) + c0*v2uo
                   Bup(iz+1,ix,iy) = Bup(iz+1,ix,iy) +  c1z*v2uo
                   Bup(iz-1,ix,iy) = Bup(iz-1,ix,iy) +  c1z*v2uo
                   Bup(iz+2,ix,iy) = Bup(iz+2,ix,iy) +  c2z*v2uo
                   Bup(iz-2,ix,iy) = Bup(iz-2,ix,iy) +  c2z*v2uo
                   Bup(iz+3,ix,iy) = Bup(iz+3,ix,iy) +  c3z*v2uo
                   Bup(iz-3,ix,iy) = Bup(iz-3,ix,iy) +  c3z*v2uo
                   Bup(iz+4,ix,iy) = Bup(iz+4,ix,iy) +  c4z*v2uo
                   Bup(iz-4,ix,iy) = Bup(iz-4,ix,iy) +  c4z*v2uo
                   Bup(iz+5,ix,iy) = Bup(iz+5,ix,iy) +  c5z*v2uo
                   Bup(iz-5,ix,iy) = Bup(iz-5,ix,iy) +  c5z*v2uo
                   Bup(iz+6,ix,iy) = Bup(iz+6,ix,iy) +  c6z*v2uo
                   Bup(iz-6,ix,iy) = Bup(iz-6,ix,iy) +  c6z*v2uo
                   Bup(iz+7,ix,iy) = Bup(iz+7,ix,iy) +  c7z*v2uo
                   Bup(iz-7,ix,iy) = Bup(iz-7,ix,iy) +  c7z*v2uo
                   Bup(iz+8,ix,iy) = Bup(iz+8,ix,iy) +  c8z*v2uo
                   Bup(iz-8,ix,iy) = Bup(iz-8,ix,iy) +  c8z*v2uo
                 
                   Bup(iz,ix+1,iy) = Bup(iz,ix+1,iy) +  c1x*v2uo
                   Bup(iz,ix-1,iy) = Bup(iz,ix-1,iy) +  c1x*v2uo
                   Bup(iz,ix+2,iy) = Bup(iz,ix+2,iy) +  c2x*v2uo
                   Bup(iz,ix-2,iy) = Bup(iz,ix-2,iy) +  c2x*v2uo
                   Bup(iz,ix+3,iy) = Bup(iz,ix+3,iy) +  c3x*v2uo
                   Bup(iz,ix-3,iy) = Bup(iz,ix-3,iy) +  c3x*v2uo
                   Bup(iz,ix+4,iy) = Bup(iz,ix+4,iy) +  c4x*v2uo
                   Bup(iz,ix-4,iy) = Bup(iz,ix-4,iy) +  c4x*v2uo
                   Bup(iz,ix+5,iy) = Bup(iz,ix+5,iy) +  c5x*v2uo
                   Bup(iz,ix-5,iy) = Bup(iz,ix-5,iy) +  c5x*v2uo
                   Bup(iz,ix+6,iy) = Bup(iz,ix+6,iy) +  c6x*v2uo
                   Bup(iz,ix-6,iy) = Bup(iz,ix-6,iy) +  c6x*v2uo
                   Bup(iz,ix+7,iy) = Bup(iz,ix+7,iy) +  c7x*v2uo
                   Bup(iz,ix-7,iy) = Bup(iz,ix-7,iy) +  c7x*v2uo
                   Bup(iz,ix+8,iy) = Bup(iz,ix+8,iy) +  c8x*v2uo
                   Bup(iz,ix-8,iy) = Bup(iz,ix-8,iy) +  c8x*v2uo

                   Bup(iz,ix,iy+1) = Bup(iz,ix,iy+1) +  c1y*v2uo
                   Bup(iz,ix,iy-1) = Bup(iz,ix,iy-1) +  c1y*v2uo
                   Bup(iz,ix,iy+2) = Bup(iz,ix,iy+2) +  c2y*v2uo
                   Bup(iz,ix,iy-2) = Bup(iz,ix,iy-2) +  c2y*v2uo
                   Bup(iz,ix,iy+3) = Bup(iz,ix,iy+3) +  c3y*v2uo
                   Bup(iz,ix,iy-3) = Bup(iz,ix,iy-3) +  c3y*v2uo
                   Bup(iz,ix,iy+4) = Bup(iz,ix,iy+4) +  c4y*v2uo
                   Bup(iz,ix,iy-4) = Bup(iz,ix,iy-4) +  c4y*v2uo
                   Bup(iz,ix,iy+5) = Bup(iz,ix,iy+5) +  c5y*v2uo
                   Bup(iz,ix,iy-5) = Bup(iz,ix,iy-5) +  c5y*v2uo
                   Bup(iz,ix,iy+6) = Bup(iz,ix,iy+6) +  c6y*v2uo
                   Bup(iz,ix,iy-6) = Bup(iz,ix,iy-6) +  c6y*v2uo
                   Bup(iz,ix,iy+7) = Bup(iz,ix,iy+7) +  c7y*v2uo
                   Bup(iz,ix,iy-7) = Bup(iz,ix,iy-7) +  c7y*v2uo
                   Bup(iz,ix,iy+8) = Bup(iz,ix,iy+8) +  c8y*v2uo
                   Bup(iz,ix,iy-8) = Bup(iz,ix,iy-8) +  c8y*v2uo

                  end do
                end do
              end do
 

             !$OMP CRITICAL
               do iy = 1,ydbloc+2*nop
                  jy = iy+yib-1-nop                   
                 do ix = 1,xdbloc+2*nop
                    jx = ix+xib-1-nop
                   do iz = 1,zdbloc+2*nop
                      jz = iz+zib-1-nop
                      up(jz,jx,jy) = up(jz,jx,jy) + Bup(iz,ix,iy)
                   end do
                 end do
               end do
             !$OMP END CRITICAL

	    end do
	   end do
	 end do
	!$OMP END PARALLEL DO

    end if

  end subroutine

!====================================================================================
  subroutine Gaussian1D(adj,m,nz,nx,ny)
  ! Take a 1D Gaussian and its adjoint

  logical                   :: adj
  integer                   :: nz, nx, ny, ix, iy, iz, nop,wd, i, j, k
  real                      :: dx, dy, dz, width, val, norm , kk, width2, a
  real, dimension(:), allocatable :: gauss
  real, dimension(:,:,:)      :: m
  real, dimension(:,:,:), allocatable  :: d


  wd = 2
  norm = 0.0
  width = 0.0
  val = 0.0

  allocate(d(nz,nx,ny))
  allocate(gauss(wd+1))  

  do iy=1,ny
  do ix=1,nx
     do iz=1,nz
        d(iz,ix,iy)=0.
     end do
  end do
  end do

   write(0,*) 'Constructing filters'
  ! filter is only half of the entire coverage because it is symmetric about zero
  ! the center of the filter is gauss(1) (largest amp), and the smallest value is gauss(wd+1)
  a = 6.
  width2 = 2.0*real(wd+1)**2
  kk = a*a/width2


    do i=1,wd+1
       val = real(i-1)
       gauss(i) = exp(-kk*val*val)
       norm = norm + gauss(i)
    end do

  gauss = gauss/norm
  gauss = gauss/norm/2.0
        write(0,*) 'gauss filter',  gauss


  write(0,*) 'Applying filter'

  if (.not. adj) then
   do iy=1+wd, ny-wd  
   do ix=1+wd, nx-wd  
     do iz=1+wd, nz-wd
       do k=0,wd
          d(iz,ix,iy) = d(iz,ix,iy) + gauss(k)*(m(iz+k,ix,iy) + m(iz-k,ix,iy))      
       end do
     enddo
   end do
   end do


  else
   do iy=1+wd, ny-wd  
   do ix=1+wd, nx-wd  
     do iz=1+wd, nz-wd
       do k=0,wd
           d(iz+k,ix,iy) = d(iz+k,ix,iy) + gauss(k)*m(iz,ix,iy)
           d(iz-k,ix,iy) = d(iz-k,ix,iy) + gauss(k)*m(iz,ix,iy)    
       end do
     enddo
   end do
   end do
  end if

    write(0,*) 'm, d', m(iz/6,ix/2,iy/2), d(iz/6,ix/2,iy/2)


    do iy=1,ny
    do ix=1,nx
     do iz=1,nz
        m(iz,ix,iy)=m(iz,ix,iy) - d(iz,ix,iy)
     end do
    end do
    end do


  write(0,*) 'finish Gaussian 1D'

  end subroutine
!====================================================================================
end module
